;; Generated package description from scanner.el  -*- no-byte-compile: t -*-
(define-package "scanner" "0.1.0.20200410.201652" "Scan documents and images" '((emacs "25.1") (dash "2.12.0")) :keywords '("hardware" "multimedia") :authors '(("Raffael Stocker" . "r.stocker@mnet-mail.de")) :maintainer '("Raffael Stocker" . "r.stocker@mnet-mail.de") :url "https://gitlab.com/rstocker/scanner.git")
